import torch
import torch.nn as nn
import torch.nn.functional as F
from .emd_utils import *
from .resnet import ResNet


class DeepEMD(nn.Module):

    def __init__(self, mode='meta'):
        super().__init__()

        self.mode = mode
        self.args = None
        self.solver = 'opencv'
        self.temperature = 12.5
        self.form = None
        #self.l2_strength = ''
        self.norm = 'center'
        self.metric = 'cosine'

        #self.encoder = ResNet(args=args)

        if self.mode == 'pre_train':
            self.fc = nn.Linear(640, self.args.num_class)

    def forward(self, support, query):
        if self.mode == 'meta':
            # support, query = input
            return self.emd_forward_1shot(support, query)

        elif self.mode == 'pre_train':
            return self.pre_train_forward(input)

        elif self.mode == 'encoder':
            if self.args.deepemd == 'fcn':
                dense = True
            else:
                dense = False
            return self.encode(input, dense)
        else:
            raise ValueError('Unknown mode')

    def pre_train_forward(self, input):
        return self.fc(self.encode(input, dense=False).squeeze(-1).squeeze(-1))

    def get_weight_vector(self, A, B):
        # A[75, 640, 5, 5], B[5, 640, 5, 5]
        M = A.shape[0]
        N = B.shape[0]

        B = F.adaptive_avg_pool2d(B, [1, 1])  # [5, 640, 1, 1]
        B = B.repeat(1, 1, A.shape[2], A.shape[3])  # [5, 640, 5, 5]

        A = A.unsqueeze(1)  # [75, 1, 640, 5, 5]
        B = B.unsqueeze(0)  # [1, 5, 640, 5, 5]

        A = A.repeat(1, N, 1, 1, 1)  # [75, 5, 640, 5, 5]
        B = B.repeat(M, 1, 1, 1, 1)  # [75, 5, 640, 5, 5]

        combination = (A * B).sum(2)  # (A * B): [75, 5, 640, 5, 5] * [75, 5, 640, 5, 5] = [75, 5, 640, 5, 5], (A * B).sum(2)= [75, 5, 5, 5]
        combination = combination.view(M, N, -1)  # [75, 5, 25]
        combination = F.relu(combination) + 1e-3  # [75, 5, 25]
        return combination

    def emd_forward_1shot(self, proto, query):
        # proto [5, 640, 5, 5], query[75, 640, 5, 5]
        #proto = proto.squeeze(0)

        weight_1 = self.get_weight_vector(query, proto)  # t[75, 5, 25]
        weight_2 = self.get_weight_vector(proto, query)  # t[5, 75, 25]

        proto = self.normalize_feature(proto)  # [5, 640, 5, 5]
        query = self.normalize_feature(query)  # [75, 640, 5, 5]

        similarity_map = self.get_similiarity_map(proto, query)  # [75, 5, 25, 25]
        if self.solver == 'opencv' or (not self.training):
            logits = self.get_emd_distance(similarity_map, weight_1, weight_2, solver='opencv')
        else:
            logits = self.get_emd_distance(similarity_map, weight_1, weight_2, solver='qpth')
        return logits

    def get_sfc(self, support):
        support = support.squeeze(0)
        # init the proto
        SFC = support.view(self.args.shot, -1, 640, support.shape[-2], support.shape[-1]).mean(dim=0).clone().detach()
        SFC = nn.Parameter(SFC.detach(), requires_grad=True)

        optimizer = torch.optim.SGD([SFC], lr=self.args.sfc_lr, momentum=0.9, dampening=0.9, weight_decay=0)

        # crate label for finetune
        label_shot = torch.arange(self.args.way).repeat(self.args.shot)
        label_shot = label_shot.type(torch.cuda.LongTensor)

        with torch.enable_grad():
            for k in range(0, self.args.sfc_update_step):
                rand_id = torch.randperm(self.args.way * self.args.shot).cuda()
                for j in range(0, self.args.way * self.args.shot, self.args.sfc_bs):
                    selected_id = rand_id[j: min(j + self.args.sfc_bs, self.args.way * self.args.shot)]
                    batch_shot = support[selected_id, :]
                    batch_label = label_shot[selected_id]
                    optimizer.zero_grad()
                    logits = self.emd_forward_1shot(SFC, batch_shot.detach())
                    loss = F.cross_entropy(logits, batch_label)
                    loss.backward()
                    optimizer.step()
        return SFC

    def get_emd_distance(self, similarity_map, weight_1, weight_2, solver='opencv'):
        num_query = similarity_map.shape[0]
        num_proto = similarity_map.shape[1]
        num_node = weight_1.shape[-1]  # 25

        if solver == 'opencv':  # use openCV solver

            for i in range(num_query):
                for j in range(num_proto):
                    # 1 - similarity_map[i, j, :, :] : [25, 25]  weight_1[i, j, :]: 25
                    _, flow = emd_inference_opencv(1 - similarity_map[i, j, :, :], weight_1[i, j, :], weight_2[j, i, :])  # flow [25 , 25]

                    similarity_map[i, j, :, :] = (similarity_map[i, j, :, :])*torch.from_numpy(flow).cuda()  #

            temperature = (self.temperature/num_node)  # similarity_map: t[75, 5, 25, 25], similarity_map.sum(-1).sum(-1): [75, 5]
            logitis = similarity_map.sum(-1).sum(-1) * temperature
            return logitis  # [query_num, proto_num]

        elif solver == 'qpth':
            weight_2 = weight_2.permute(1, 0, 2)
            similarity_map = similarity_map.view(num_query * num_proto, similarity_map.shape[-2],
                                                 similarity_map.shape[-1])
            weight_1 = weight_1.view(num_query * num_proto, weight_1.shape[-1])
            weight_2 = weight_2.reshape(num_query * num_proto, weight_2.shape[-1])

            _, flows = emd_inference_qpth(1 - similarity_map, weight_1, weight_2,form=self.form, l2_strength=self.l2_strength)

            logitis=(flows*similarity_map).view(num_query, num_proto,flows.shape[-2],flows.shape[-1])
            temperature = (self.temperature / num_node)
            logitis = logitis.sum(-1).sum(-1) *  temperature
        else:
            raise ValueError('Unknown Solver')

        return logitis

    def normalize_feature(self, x):
        if self.norm == 'center':
            x = x - x.mean(1).unsqueeze(1)
            return x
        else:
            return x


    def get_similiarity_map(self, proto, query):
        way = proto.shape[0]
        num_query = query.shape[0]
        query = query.view(query.shape[0], query.shape[1], -1)  # [75, 640, 25]
        proto = proto.view(proto.shape[0], proto.shape[1], -1)  # [5, 640, 25]

        proto = proto.unsqueeze(0).repeat([num_query, 1, 1, 1])  # [75, 5, 640, 25]
        query = query.unsqueeze(1).repeat([1, way, 1, 1])  # [75, 5, 640, 25]
        proto = proto.permute(0, 1, 3, 2)  # [75, 5, 25, 640]
        query = query.permute(0, 1, 3, 2)  # [75, 5, 25, 640]
        feature_size = proto.shape[-2]  # 25

        if self.metric == 'cosine':
            proto = proto.unsqueeze(-3)  # [75, 5, 1, 25, 640]
            query = query.unsqueeze(-2)  # [75, 5, 25, 1, 640]
            query = query.repeat(1, 1, 1, feature_size, 1)  # [75, 5, 25, 25, 640]
            similarity_map = F.cosine_similarity(proto, query, dim=-1)  # proto [75, 5, 1, 25, 640] , query [75, 5, 25, 25, 640]  ->  [75, 5, 25, 25]
        if self.metric == 'l2':
            proto = proto.unsqueeze(-3)
            query = query.unsqueeze(-2)
            query = query.repeat(1, 1, 1, feature_size, 1)
            similarity_map = (proto - query).pow(2).sum(-1)
            similarity_map = 1 - similarity_map

        return similarity_map

    def encode(self, x, dense=True):

        if x.shape.__len__() == 5:  # batch of image patches
            num_data, num_patch = x.shape[:2]
            x = x.reshape(-1, x.shape[2], x.shape[3], x.shape[4])
            x = self.encoder(x)
            x = F.adaptive_avg_pool2d(x, 1)
            x = x.reshape(num_data, num_patch, x.shape[1], x.shape[2], x.shape[3])
            x = x.permute(0, 2, 1, 3, 4)
            x = x.squeeze(-1)
            return x

        else:
            x = self.encoder(x)
            if dense == False:
                x = F.adaptive_avg_pool2d(x, 1)
                return x
            if self.args.feature_pyramid is not None:
                x = self.build_feature_pyramid(x)
        return x

    def build_feature_pyramid(self, feature):
        feature_list = []
        for size in self.args.feature_pyramid:
            feature_list.append(F.adaptive_avg_pool2d(feature, size).view(feature.shape[0], feature.shape[1], 1, -1))
        feature_list.append(feature.view(feature.shape[0], feature.shape[1], 1, -1))
        out = torch.cat(feature_list, dim=-1)
        return out
