import torch
import torch.nn as nn
import torch.nn.functional as F
from .emd_utils import *
from .resnet import ResNet


class EMD_weight_generator(nn.Module):

    def __init__(self ):
        super().__init__()

        self.args = None
        self.solver = 'opencv'
        self.temperature = 12.5
        self.form = None
        self.norm = 'center'
        self.metric = 'cosine'

    def forward(self, support, query):
            return self.emd_forward(support, query)

    def get_weight_vector(self, A, B):
        # A[75, 640, 5, 5], B[5, 640, 5, 5]
        M = A.shape[0]
        N = B.shape[0]

        B = F.adaptive_avg_pool2d(B, [1, 1])  # [5, 640, 1, 1]
        B = B.repeat(1, 1, A.shape[2], A.shape[3])  # [5, 640, 5, 5]

        A = A.unsqueeze(1)  # [n, 1, 128, 6, 6]
        B = B.unsqueeze(0)  # [1, n, 128, 6, 6]

        A = A.repeat(1, N, 1, 1, 1)  # [75, 5, 640, 5, 5]
        B = B.repeat(M, 1, 1, 1, 1)  # [75, 5, 640, 5, 5]

        combination = (A * B).sum(2)  # (A * B): [75, 5, 640, 5, 5] * [75, 5, 640, 5, 5] = [75, 5, 640, 5, 5], (A * B).sum(2)= [75, 5, 5, 5]
        combination = combination.view(M, N, -1)  # [75, 5, 25]
        combination = F.relu(combination) + 1e-3  # [75, 5, 25]
        return combination

    def emd_forward(self, proto, query):
        # proto [5, 640, 5, 5], query[75, 640, 5, 5]
        #proto = proto.squeeze(0)

        weight_1 = self.get_weight_vector(query, proto)  # t[1, 1, 36]
        weight_2 = self.get_weight_vector(proto, query)  # t[1, 1, 36]



        proto = self.normalize_feature(proto)  # [5, 640, 5, 5]
        query = self.normalize_feature(query)  # [75, 640, 5, 5]

        similarity_map = self.get_similiarity_map(proto, query)  # [75, 5, 25, 25]
        if self.solver == 'opencv' or (not self.training):
            logits = self.get_emd_distance(similarity_map, weight_1, weight_2, solver='opencv')
        else:
            logits = self.get_emd_distance(similarity_map, weight_1, weight_2, solver='qpth')
        return logits

    def get_emd_distance(self, similarity_map, weight_1, weight_2, solver='opencv'):
        num_query = similarity_map.shape[0]
        num_proto = similarity_map.shape[1]
        num_node = weight_1.shape[-1]  # 25

        if solver == 'opencv':  # use openCV solver

            for i in range(num_query):
                for j in range(num_proto):
                    # 1 - similarity_map[i, j, :, :] : [25, 25]  weight_1[i, j, :]: 25
                    _, flow = emd_inference_opencv(1 - similarity_map[i, j, :, :], weight_1[i, j, :], weight_2[j, i, :])  # flow [25 , 25]

                    similarity_map[i, j, :, :] = (similarity_map[i, j, :, :])*torch.from_numpy(flow).cuda()  #

            temperature = (self.temperature/num_node)  # similarity_map: t[75, 5, 25, 25], similarity_map.sum(-1).sum(-1): [75, 5]
            logitis = similarity_map.sum(-1).sum(-1) * temperature
            return logitis  # [query_num, proto_num]

        elif solver == 'qpth':
            weight_2 = weight_2.permute(1, 0, 2)
            similarity_map = similarity_map.view(num_query * num_proto, similarity_map.shape[-2],
                                                 similarity_map.shape[-1])
            weight_1 = weight_1.view(num_query * num_proto, weight_1.shape[-1])
            weight_2 = weight_2.reshape(num_query * num_proto, weight_2.shape[-1])

            _, flows = emd_inference_qpth(1 - similarity_map, weight_1, weight_2,form=self.form, l2_strength=self.l2_strength)

            logitis=(flows*similarity_map).view(num_query, num_proto,flows.shape[-2],flows.shape[-1])
            temperature = (self.temperature / num_node)
            logitis = logitis.sum(-1).sum(-1) *  temperature
        else:
            raise ValueError('Unknown Solver')

        return logitis

    def normalize_feature(self, x):
        if self.norm == 'center':
            x = x - x.mean(1).unsqueeze(1)
            return x
        else:
            return x

    def get_similiarity_map(self, proto, query):
        way = proto.shape[0]
        num_query = query.shape[0]
        query = query.view(query.shape[0], query.shape[1], -1)  # [75, 640, 25]
        proto = proto.view(proto.shape[0], proto.shape[1], -1)  # [5, 640, 25]

        proto = proto.unsqueeze(0).repeat([num_query, 1, 1, 1])  # [75, 5, 640, 25]
        query = query.unsqueeze(1).repeat([1, way, 1, 1])  # [75, 5, 640, 25]
        proto = proto.permute(0, 1, 3, 2)  # [75, 5, 25, 640]
        query = query.permute(0, 1, 3, 2)  # [75, 5, 25, 640]
        feature_size = proto.shape[-2]  # 25

        if self.metric == 'cosine':
            proto = proto.unsqueeze(-3)  # [75, 5, 1, 25, 640]
            query = query.unsqueeze(-2)  # [75, 5, 25, 1, 640]
            query = query.repeat(1, 1, 1, feature_size, 1)  # [75, 5, 25, 25, 640]
            similarity_map = F.cosine_similarity(proto, query, dim=-1)  # proto [75, 5, 1, 25, 640] , query [75, 5, 25, 25, 640]  ->  [75, 5, 25, 25]
        if self.metric == 'l2':
            proto = proto.unsqueeze(-3)
            query = query.unsqueeze(-2)
            query = query.repeat(1, 1, 1, feature_size, 1)
            similarity_map = (proto - query).pow(2).sum(-1)
            similarity_map = 1 - similarity_map

        return similarity_map

