import pytracking.evaluation.vot2020 as vot
import importlib
import os
import cv2
import numpy as np
import time
import torch

import seaborn as sns
import matplotlib.pyplot as plt

def get_parameters(tracker_name, param_file_name):
    param_module = importlib.import_module('pytracking.parameter.{}.{}'.format(tracker_name, param_file_name))
    params = param_module.parameters()
    return params

def read_image( image_file: str):
    im = cv2.imread(image_file)
    return cv2.cvtColor(im, cv2.COLOR_BGR2RGB)


def _convert_anno_to_list(vot_anno):
    vot_anno = [vot_anno[0], vot_anno[1], vot_anno[2], vot_anno[3]]
    return vot_anno

def _convert_image_path(image_path):
    return image_path

def masks_to_bboxes(mask, fmt='c'):

    """ Convert a mask tensor to one or more bounding boxes.
    Note: This function is a bit new, make sure it does what it says.  /Andreas
    :param mask: Tensor of masks, shape = (..., H, W)
    :param fmt: bbox layout. 'c' => "center + size" or (x_center, y_center, width, height)
                             't' => "top left + size" or (x_left, y_top, width, height)
                             'v' => "vertices" or (x_left, y_top, x_right, y_bottom)
    :return: tensor containing a batch of bounding boxes, shape = (..., 4)
    """
    batch_shape = mask.shape[:-2]
    mask = mask.reshape((-1, *mask.shape[-2:]))
    bboxes = []

    for m in mask:
        mx = m.sum(dim=-2).nonzero()
        my = m.sum(dim=-1).nonzero()
        bb = [mx.min(), my.min(), mx.max(), my.max()] if (len(mx) > 0 and len(my) > 0) else [0, 0, 0, 0]
        bboxes.append(bb)

    bboxes = torch.tensor(bboxes, dtype=torch.float32, device=mask.device)
    bboxes = bboxes.reshape(batch_shape + (4,))

    if fmt == 'v':
        return bboxes

    x1 = bboxes[..., :2]
    s = bboxes[..., 2:] - x1 + 1

    if fmt == 'c':
        return torch.cat((x1 + 0.5 * s, s), dim=-1)
    elif fmt == 't':
        return torch.cat((x1, s), dim=-1)

    raise ValueError("Undefined bounding box layout '%s'" % fmt)

def run_vot_2020():
    torch.backends.cudnn.benchmark = False

    tracker_module_DMB = importlib.import_module('pytracking.tracker.{}'.format('DMB_506_tracker'))
    tracker_class_DMB = tracker_module_DMB.get_tracker_class()
    param_module = importlib.import_module('pytracking.parameter.{}.{}'.format('DMB', 'DMB_506'))
    params = param_module.parameters()
    DMB_train_tracker = tracker_class_DMB(params)

    tracker = DMB_train_tracker

    output_segmentation = True

    if output_segmentation:
        handle = vot.VOT("mask")
    else:
        handle = vot.VOT("rectangle")

    vot_anno = handle.region()

    image_path = handle.frame()
    if not image_path:
        return
    image_path = _convert_image_path(image_path)

    image = read_image(image_path)

    if output_segmentation:
        vot_anno_mask = vot.make_full_size(vot_anno, (image.shape[1], image.shape[0]))
        bbox = masks_to_bboxes(torch.from_numpy(vot_anno_mask), fmt='t').squeeze().tolist()
    # image, state, init_mask=None

    tracker.initialize(image, bbox, vot_anno_mask)

    while True:
        image_path = handle.frame()
        if not image_path:
            break
        image_path = _convert_image_path(image_path)

        image = read_image(image_path)

        out = tracker.track_mask(image)

        if output_segmentation:
            pred = out
            #import seaborn as sns

            #sns.heatmap(pred)
            #plt.show()

        handle.report(pred, 1.0)

if __name__ == '__main__':
    run_vot_2020()


