# -*- coding: utf-8 -*
import math
import os
from os import makedirs
from os.path import isdir, join

import cv2
import numpy as np
from loguru import logger
from PIL import Image
from tqdm import tqdm

import torch
import torch.multiprocessing as mp
from multiprocessing import Process, Queue

from torch.multiprocessing import Manager
import torch.nn.functional as F
import os
import os.path as osp

from DavisTester.eval import davis_benchmark


def ensure_dir(dir_path: str):
    r"""
    Ensure the existence of path (i.e. mkdir -p)
    Arguments
    ---------
    dir_path: str
        path to be ensured
    """
    if osp.exists(dir_path):
        return
    else:
        os.makedirs(dir_path)


class DAVISTester:
    _hyper_params = dict(
          exp_name="d3s_seg",
          exp_save="/home/jaffe/PycharmProjects/d3s_me/aaai/segm/stm_train_davis",
          device_num=1,
          data_root="/home/jaffe/Titanspace/datasets/dataset/testdataset/dataset/Davis/DAVIS2017",
          dataset_names=[],
          save_video=False,
          save_patch=False
    )

    def __init__(self, segmenter, dataset=None):
        super(DAVISTester, self).__init__()
        self._state = dict()  # pipeline state
        self._state['speed'] = -1
        self.iou_eval_thres = np.arange(0.3, 0.5, 0.05)
        self._pipeline = segmenter
        if dataset is not None:
           self._hyper_params['dataset_names'].append(dataset)


    def test(self):
        r"""
        Run test
        """
        # set dir
        self.tracker_name = self._hyper_params["exp_name"]
        for dataset_name in self._hyper_params["dataset_names"]:
            self.dataset_name = dataset_name
            self.tracker_dir = os.path.join(self._hyper_params["exp_save"],self.dataset_name)
            self.save_root_dir = os.path.join(self.tracker_dir, self.tracker_name, "baseline")
            ensure_dir(self.save_root_dir)
            # track videos
            self.run_tracker()
            # evaluation
            eval_result = self.evaluation('default_hp')
        return dict(main_performance=eval_result["JF"])

    def run_tracker(self):
        """
        Run self.pipeline on DAVIS
        """
        num_gpu = self._hyper_params["device_num"]
        all_devs = [torch.device("cuda:%d" % i) for i in range(num_gpu)]
        logger.info('runing test on devices {}'.format(all_devs))
        davis_root = self._hyper_params["data_root"]
        logger.info('Using dataset %s at: %s' % (self.dataset_name, davis_root))
        # setup dataset
        dataset = davis_benchmark.load_dataset(davis_root, self.dataset_name) # OrderedDict:30 bike-packing, blackswan, bmx-trees...
        self.dataset = dataset
        keys = list(dataset.keys()) # list:30
        keys.sort()
        nr_records = len(keys)
        pbar = tqdm(total=nr_records)
        mean_speed = -1
        speed_list = []
        speed_queue = Queue(500)
        # set worker
        self.worker(keys, self.dataset, speed_queue)
        for i in range(nr_records):
            s = speed_queue.get()
            speed_list.append(s)
            pbar.update(1)
        # print result
        mean_speed = float(np.mean(speed_list))
        logger.info('Mean Speed: {:.2f} FPS'.format(mean_speed))
        self._state['speed'] = mean_speed

    def worker(self, records, dataset, speed_queue=None):
        tracker = self._pipeline
        for v_id, video in enumerate(records):
            speed = self.track_single_video_vos(tracker, dataset[video])
            if speed_queue is not None:
                speed_queue.put_nowait(speed)

    def evaluation(self, search_task_name):
        r"""
        Run evaluation & write result to csv file under self.tracker_dir
        """

        results_path = join(self.save_root_dir, 'results_multi')
        davis_data_path = self._hyper_params["data_root"]

        eval_dump_path = join(self.save_root_dir, 'dump')
        if not isdir(eval_dump_path): makedirs(eval_dump_path)

        csv_name_global_path = join(eval_dump_path,
                                    search_task_name + '_global_results.csv')
        csv_name_per_sequence_path = join(
            eval_dump_path, search_task_name + '_name_per_sequence_results.csv')

        version = self.dataset_name[-4:]
        hp_dict = {}
        return davis_benchmark.davis2017_eval(davis_data_path,
                                              results_path,
                                              csv_name_global_path,
                                              csv_name_per_sequence_path,
                                              hp_dict,
                                              version=version)
    # video = dict[3]: anno_files, image_files, name
    def Soft_aggregation(self, ps, K):
        def ToCuda(xs, cuda_index=0):
            if torch.cuda.is_available():
                if isinstance(xs, list) or isinstance(xs, tuple):
                    return [x.cuda(cuda_index) for x in xs]
                else:
                    return xs.cuda(cuda_index)
            else:
                return xs

        num_objects, H, W = ps.shape
        em = ToCuda(torch.zeros(1, K, H, W), ps.device.index)
        em[0, 0] = torch.prod(1 - ps, dim=0)  # bg prob
        em[0, 1:num_objects + 1] = ps  # obj
        em = torch.clamp(em, 1e-7, 1 - 1e-7)
        logit = torch.log((em / (1 - em)))
        return logit

    def pad_divide_by(self, d, in_size):
        out_list = []
        h, w = in_size
        if h % d > 0:
            new_h = h + d - h % d
        else:
            new_h = h
        if w % d > 0:
            new_w = w + d - w % d
        else:
            new_w = w
        lh, uh = int((new_h - h) / 2), int(new_h - h) - int((new_h - h) / 2)
        lw, uw = int((new_w - w) / 2), int(new_w - w) - int((new_w - w) / 2)
        pad_array = (int(lw), int(uw), int(lh), int(uh))

        return pad_array

    def track_single_video_vos(self, tracker, video, mot_enable=True):
        '''
        perfrom semi-supervised video object segmentation for single video
        :param tracker: tracker pipeline
        :param video: video info
        :param mot_enable:  if true, perform instance level segmentation on davis, otherwise semantic
        '''

        if 'DAVIS' in self._hyper_params["data_root"]:
            self.palette = Image.open(self._hyper_params["data_root"] + '/Annotations/480p/'+ video['name'] +'/00000.png').getpalette()

        if 'tianchi' in self._hyper_params["data_root"]:
            self.palette = Image.open(self._hyper_params["data_root"] + '/Annotations/'+ video['name'] +'/00000.png').getpalette()

        image_files = video['image_files']

        if '16' in self._hyper_params["dataset_names"][0]:
            annos = [np.array(Image.open(x))/255 for x in video['anno_files']]
        else:
            annos = [np.array(Image.open(x)) for x in video['anno_files']]

        if 'anno_init_files' in video:
            annos_init = [
                np.array(Image.open(x)) for x in video['anno_init_files']
            ]
        else:
            annos_init = [annos[0]] # mask np.array[h, w] :0, 1...

        if not mot_enable:
            annos = [(anno > 0).astype(np.uint8) for anno in annos]
            annos_init = [(anno_init > 0).astype(np.uint8)
                          for anno_init in annos_init]

        if 'start_frame' in video:
            object_ids = [int(id) for id in video['start_frame']]
        else:
            object_ids = [o_id for o_id in np.unique(annos[0]) if o_id != 0] # 视频有几个 要分割 的物体
            if len(object_ids) != len(annos_init):
                annos_init = annos_init * len(object_ids) # list[2]
        object_num = len(object_ids)
        toc = 0
        pred_masks = np.zeros((object_num, len(image_files), annos[0].shape[0],
                               annos[0].shape[1])) - 1 # 所有图片的GT np.array[object_num, num_imgs, h, w]
        if self._hyper_params['save_video']:
            track_boxes = np.zeros((object_num, len(image_files), 4))
            track_mask_boxes = np.zeros((object_num, len(image_files), 4))
            track_mask_score = np.zeros((object_num, len(image_files)))
            track_score = np.zeros((object_num, len(image_files)))
            state_score = np.zeros((object_num, len(image_files)))
        if self._hyper_params['save_patch']:
            patch_list = []

        for obj_id, o_id in enumerate(object_ids): # list[object_num], [1, 2, ...]
            obj_patch_list = []
            logger.info('{} th object in video {}'.format(o_id, video['name']))
            if 'start_frame' in video:
                start_frame = video['start_frame'][str(o_id)]
                end_frame = video['end_frame'][str(o_id)]
            else:
                start_frame, end_frame = 0, len(image_files)

            for f, image_file in enumerate(tqdm(image_files)): # image_files list[num_images], 图片路径
               # im = cv2.imread(image_file)
                im = cv2.cvtColor(cv2.imread(image_file), cv2.COLOR_BGR2RGB)
                img_h, img_w = im.shape[0], im.shape[1]
                #cv2.imwrite('/data2/jaffeProj/debug/input/' + video['name'] + str(f) +'.png', im)

                tic = cv2.getTickCount()
                if f == start_frame:  # init
                    mask = (annos_init[obj_id] == o_id).astype(np.uint8) # np.array[h, w]: 0, 1
                    x, y, w, h = cv2.boundingRect((mask).astype(np.uint8))
                    tracker.init(im, np.array([x, y, w, h]), video = video['name'] , init_mask = mask)
                elif end_frame >= f > start_frame:  # tracking
                    mask = tracker.update_vos(im) # np.array[h, w]

                toc += cv2.getTickCount() - tic
                if end_frame >= f >= start_frame:
                    pred_masks[obj_id, f, :, :] = mask #[num_ob, frame, h, w]

        toc /= cv2.getTickFrequency()

        video_path = join(self.save_root_dir, 'results_multi', video['name'])
        logger.info('save mask path:{}'.format(video_path))
        if not isdir(video_path): makedirs(video_path)

        pred_mask_final = np.array(pred_masks)  # [num_ob, frame, h, w]
        pred_mask_final = (np.argmax(pred_mask_final, axis=0).astype('uint8') + 1) * (np.max(pred_mask_final, axis=0) > 0.35).astype('uint8')

        # [frame, h, w]
        for i in range(pred_mask_final.shape[0]):
            mask_label = pred_mask_final[i].astype(np.uint8)
            # sns.heatmap(mask_label)
            # plt.show()
            mask_label = Image.fromarray(mask_label)
            mask_label.putpalette(self.palette)
            mask_label.save(join(video_path, image_files[i].split('/')[-1].split('.')[0] + '.png'))



        logger.info('({:d}) Video: {:12s} Time: {:02.1f}s Speed: {:3.1f}fps'.format(int(o_id), video['name'], toc,f * len(object_ids) / toc))
        speed = f * len(object_ids) / toc
        logger.info("{} speed: {}".format(video['name'], speed))

        return speed


